diff mbox

[V2] mtd: block2mtd: Present block2mtd timely on boot time

Message ID 1807144344.40128259.1410985683342.JavaMail.zimbra@redhat.com
State Changes Requested
Headers show

Commit Message

Rodrigo Freire Sept. 17, 2014, 8:28 p.m. UTC
From: Felix Fietkau <nbd@openwrt.org>

mtd: block2mtd: Ensure that block2mtd is presented in a timely fashion

Currently, a block MTD device is not presented to the system on time, in
order to start mounting the filesystems. This patch ensures that block2mtd
is presented at the right time, so filesystems can be mounted on boot time.
This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
block2mtd filesystems.
This patchset also adds a MTD device name and a timeout option to the driver.
Original patchset:
https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/440-block2mtd_init.patch?rev=40444
https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/441-block2mtd_probe.patch?rev=40444

Signed-off-by: Felix Fietkau <nbd@openwrt.org>
Signed-off-by: Rodrigo Freire <rfreire@redhat.com>
Signed-off-by: Herton Krzesinski <herton@redhat.com>
---
V2: Uses kstrdup, removed PAGE_MASK.
---
1.7.1

Comments

Ezequiel Garcia Sept. 17, 2014, 9:21 p.m. UTC | #1
On 17 September 2014 21:28, Rodrigo Freire <rfreire@redhat.com> wrote:
>
> mtd: block2mtd: Ensure that block2mtd is presented in a timely fashion
>

Using block2mtd sounds a bit unusual. I see that you are trying to get
a more robust fs.... have you tried using f2fs instead of jffs2?

> Currently, a block MTD device is not presented to the system on time, in
> order to start mounting the filesystems. This patch ensures that block2mtd
> is presented at the right time, so filesystems can be mounted on boot time.

It worries me a bit to add such a long delay to the boot. If for some
reason the SD is not working, then the kernel will wait (by default) 3
seconds now?
Rodrigo Freire Sept. 17, 2014, 9:41 p.m. UTC | #2
Holas Ezequiel,

----- Original Message ----- 
From: "Ezequiel Garcia" <ezequiel@vanguardiasur.com.ar> 
> On 17 September 2014 21:28, Rodrigo Freire <rfreire@redhat.com> wrote: 
> 
> Using block2mtd sounds a bit unusual. I see that you are trying to get 
> a more robust fs.... have you tried using f2fs instead of jffs2? 

I see that it is still marked as Experimental as of latest (3.17-RC5)
kernel. But will take a look in the future, thanks for pointing.


> > Currently, a block MTD device is not presented to the system on time, in 
> > order to start mounting the filesystems. This patch ensures that block2mtd 
> > is presented at the right time, so filesystems can be mounted on boot time. 
> 
> It worries me a bit to add such a long delay to the boot. If for some 
> reason the SD is not working, then the kernel will wait (by default) 3 
> seconds now? 

Not really; see the decision path:
IF block2mtd is not a module (is builtin), AND
   IF there is a valid block2mtd= clause on kernel cmdline, AND
      IF the device specified device on block2mtd= clause is still not present
      THEN wait *up to* 3 seconds (or seconds=n if specified on block2mtd=
           cmdline) to the device to show up. If the device shows up earlier,
           the device is created and boot proceeds.
      ELSE Fail to create the block2mtd device.
ELSE keep booting the kernel normally, without any further delays.

Best regards,

- RF.
Brian Norris Nov. 5, 2014, 8:23 p.m. UTC | #3
On Wed, Sep 17, 2014 at 04:28:03PM -0400, Rodrigo Freire wrote:
> From: Felix Fietkau <nbd@openwrt.org>
> 
> mtd: block2mtd: Ensure that block2mtd is presented in a timely fashion
> 
> Currently, a block MTD device is not presented to the system on time, in
> order to start mounting the filesystems. This patch ensures that block2mtd
> is presented at the right time, so filesystems can be mounted on boot time.
> This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
> block2mtd filesystems.

This still seems like a bad idea (using a block device + block2mtd +
JFFS2). Why are you doing this? See comments here:

  http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2

> This patchset also adds a MTD device name and a timeout option to the driver.
> Original patchset:
> https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/440-block2mtd_init.patch?rev=40444
> https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/441-block2mtd_probe.patch?rev=40444

You're stating right up front that this patch is doing several different
things. Please split these up into separate commits which get their own
description.

> Signed-off-by: Felix Fietkau <nbd@openwrt.org>
> Signed-off-by: Rodrigo Freire <rfreire@redhat.com>
> Signed-off-by: Herton Krzesinski <herton@redhat.com>
> ---
> V2: Uses kstrdup, removed PAGE_MASK.

You have several checkpatch warnings. Please fix them.

WARNING:LONG_LINE: line over 80 characters
#57: FILE: drivers/mtd/devices/block2mtd.c:221:
+static struct block2mtd_dev *add_device(char *devname, int erase_size, const char *mtdname, int timeout)

WARNING:SPACING: space prohibited between function name and open parenthesis '('
#111: FILE: drivers/mtd/devices/block2mtd.c:286:
+	name = kstrdup (mtdname, GFP_KERNEL);

WARNING:LONG_LINE: line over 80 characters
#150: FILE: drivers/mtd/devices/block2mtd.c:385:
+	char buf[80 + 12 + 80 + 8]; /* 80 for device, 12 for erase size, 80 for name, 8 for timeout */

total: 0 errors, 3 warnings, 164 lines checked

> --- a/drivers/mtd/devices/block2mtd.c	2014-09-16 21:38:12.543952627 -0300
> +++ b/drivers/mtd/devices/block2mtd.c	2014-09-17 17:43:21.424944394 -0300
> @@ -9,7 +9,15 @@
>  
>  #define pr_fmt(fmt) KBUILD_MODNAME ": " fmt
>  
> +/*
> +* When the first attempt at device initialization fails, we may need to
> +* wait a little bit and retry. This timeout, by default 3 seconds, gives
> +* device time to start up. Required on BCM2708 and a few other chipsets.
> +*/
> +#define MTD_DEFAULT_TIMEOUT	3
> +
>  #include <linux/module.h>
> +#include <linux/delay.h>
>  #include <linux/fs.h>
>  #include <linux/blkdev.h>
>  #include <linux/bio.h>
> @@ -17,6 +25,7 @@
>  #include <linux/list.h>
>  #include <linux/init.h>
>  #include <linux/mtd/mtd.h>
> +#include <linux/mtd/partitions.h>
>  #include <linux/mutex.h>
>  #include <linux/mount.h>
>  #include <linux/slab.h>
> @@ -209,12 +218,14 @@ static void block2mtd_free_device(struct
>  }
>  
>  
> -static struct block2mtd_dev *add_device(char *devname, int erase_size)
> +static struct block2mtd_dev *add_device(char *devname, int erase_size, const char *mtdname, int timeout)

The addition of this name parameter should definitely be its own patch.

>  {
>  	const fmode_t mode = FMODE_READ | FMODE_WRITE | FMODE_EXCL;
> -	struct block_device *bdev;
> +	struct block_device *bdev = ERR_PTR(-ENODEV);
>  	struct block2mtd_dev *dev;
> +	struct mtd_partition *part;
>  	char *name;
> +	int i;

This variable produces a warning when built as a module:

   drivers/mtd/devices/block2mtd.c: In function ‘add_device’:
   drivers/mtd/devices/block2mtd.c:228:6: warning: unused variable ‘i’ [-Wunused-variable]
     int i;
         ^

>  
>  	if (!devname)
>  		return NULL;
> @@ -225,15 +236,28 @@ static struct block2mtd_dev *add_device(
>  
>  	/* Get a handle on the device */
>  	bdev = blkdev_get_by_path(devname, mode, dev);
> -#ifndef MODULE
> -	if (IS_ERR(bdev)) {
>  
> -		/* We might not have rootfs mounted at this point. Try
> -		   to resolve the device name by other means. */
> -
> -		dev_t devt = name_to_dev_t(devname);
> -		if (devt)
> -			bdev = blkdev_get_by_dev(devt, mode, dev);
> +#ifndef MODULE
> +/*
> +* We might not have the root device mounted at this point.
> +* Try to resolve the device name by other means.
> +*/

These lines should start with a space, so the asterisks line up.

> +	for (i = 0; IS_ERR(bdev) && i <= timeout; i++) {
> +		dev_t devt;
> +
> +		if (i)
> +			/*
> +			 * Calling wait_for_device_probe in the first loop 
> +			 * was not enough, sleep for a bit in subsequent
> +			 * go-arounds.
> +			*/
> +			msleep(1000);
> +		wait_for_device_probe();
> +
> +		devt = name_to_dev_t(devname);
> +		if (!devt)
> +			continue;
> +		bdev = blkdev_get_by_dev(devt, mode, dev);
>  	}
>  #endif
>  
> @@ -257,13 +281,14 @@ static struct block2mtd_dev *add_device(
>  
>  	/* Setup the MTD structure */
>  	/* make the name contain the block device in */
> -	name = kasprintf(GFP_KERNEL, "block2mtd: %s", devname);
> +	if (!mtdname)
> +		mtdname = devname;
> +	name = kstrdup (mtdname, GFP_KERNEL);
>  	if (!name)
>  		goto err_destroy_mutex;
>  
>  	dev->mtd.name = name;
> -
> -	dev->mtd.size = dev->blkdev->bd_inode->i_size & PAGE_MASK;
> +	dev->mtd.size = dev->blkdev->bd_inode->i_size & ~(erase_size - 1);

This deserves its own patch, or at least some explanation of why you're
doing this. I guess you're seeing cases where the provided erasesize is
not aligned with the size of the block device?

>  	dev->mtd.erasesize = erase_size;
>  	dev->mtd.writesize = 1;
>  	dev->mtd.writebufsize = PAGE_SIZE;
> @@ -276,15 +301,19 @@ static struct block2mtd_dev *add_device(
>  	dev->mtd.priv = dev;
>  	dev->mtd.owner = THIS_MODULE;
>  
> -	if (mtd_device_register(&dev->mtd, NULL, 0)) {
> +	part = kzalloc(sizeof(struct mtd_partition), GFP_KERNEL);
> +	part->name = name;
> +	part->offset = 0;
> +	part->size = dev->mtd.size;

Why are you doing this? This also does not fit the description of this
patch. And what's wrong with using the default partitioning options?
Won't we (if not specified in some other way) default to an
unpartitioned MTD, which covers the entire device?

> +	if (mtd_device_register(&dev->mtd, part, 1)) {
>  		/* Device didn't get added, so free the entry */
>  		goto err_destroy_mutex;
>  	}
> +
>  	list_add(&dev->list, &blkmtd_device_list);
>  	pr_info("mtd%d: [%s] erase_size = %dKiB [%d]\n",
>  		dev->mtd.index,
> -		dev->mtd.name + strlen("block2mtd: "),
> -		dev->mtd.erasesize >> 10, dev->mtd.erasesize);
> +		mtdname, dev->mtd.erasesize >> 10, dev->mtd.erasesize);
>  	return dev;
>  
>  err_destroy_mutex:
> @@ -353,11 +382,12 @@ static char block2mtd_paramline[80 + 12]
>  
>  static int block2mtd_setup2(const char *val)
>  {
> -	char buf[80 + 12]; /* 80 for device, 12 for erase size */
> +	char buf[80 + 12 + 80 + 8]; /* 80 for device, 12 for erase size, 80 for name, 8 for timeout */
>  	char *str = buf;
> -	char *token[2];
> +	char *token[4];
>  	char *name;
>  	size_t erase_size = PAGE_SIZE;
> +	unsigned long timeout = MTD_DEFAULT_TIMEOUT;
>  	int i, ret;
>  
>  	if (strnlen(val, sizeof(buf)) >= sizeof(buf)) {
> @@ -368,7 +398,7 @@ static int block2mtd_setup2(const char *
>  	strcpy(str, val);
>  	kill_final_newline(str);
>  
> -	for (i = 0; i < 2; i++)
> +	for (i = 0; i < 4; i++)
>  		token[i] = strsep(&str, ",");
>  
>  	if (str) {
> @@ -395,7 +425,13 @@ static int block2mtd_setup2(const char *
>  		}
>  	}
>  
> -	add_device(name, erase_size);
> +	if (token[2] && (strlen(token[2]) + 1 > 80))
> +		pr_err("mtd device name too long");
> +
> +
> +	if (token[3] && kstrtoul(token[3], 0, &timeout))
> +		pr_err("invalid timeout");
> +	add_device(name, erase_size, token[2], timeout);
>  
>  	return 0;
>  }
> @@ -429,7 +465,7 @@ static int block2mtd_setup(const char *v
>  
>  
>  module_param_call(block2mtd, block2mtd_setup, NULL, NULL, 0200);
> -MODULE_PARM_DESC(block2mtd, "Device to use. \"block2mtd=<dev>[,<erasesize>]\"");
> +MODULE_PARM_DESC(block2mtd, "Device to use. \"block2mtd=<dev>[,<erasesize>[,<name>[,<timeout>]]]\"");
>  
>  static int __init block2mtd_init(void)
>  {
> @@ -463,8 +499,7 @@ static void block2mtd_exit(void)
>  	}
>  }
>  
> -
> -module_init(block2mtd_init);
> +late_initcall(block2mtd_init);
>  module_exit(block2mtd_exit);
>  
>  MODULE_LICENSE("GPL");

Brian
Artem Bityutskiy Nov. 7, 2014, 2:59 p.m. UTC | #4
On Wed, 2014-11-05 at 12:23 -0800, Brian Norris wrote:
> On Wed, Sep 17, 2014 at 04:28:03PM -0400, Rodrigo Freire wrote:
> > From: Felix Fietkau <nbd@openwrt.org>
> > 
> > mtd: block2mtd: Ensure that block2mtd is presented in a timely fashion
> > 
> > Currently, a block MTD device is not presented to the system on time, in
> > order to start mounting the filesystems. This patch ensures that block2mtd
> > is presented at the right time, so filesystems can be mounted on boot time.
> > This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
> > block2mtd filesystems.
> 
> This still seems like a bad idea (using a block device + block2mtd +
> JFFS2). Why are you doing this? See comments here:
> 
>   http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2

At old days it was impossible to mount character devices, only block
devices. So people ware using mtdblock to mount jffs2. This is just a
work-around.

I did not look at this code for long time, and may have forgotten
something, but I believe you do not need mtdblock anymore for this. You
just mount the mtd device.

This driver should not be used for this.

The only usage of this driver is emulating a block device on top of NOR
flash, and in most cases, only for debugging / research purposes. This
is because (a) this driver does not handle bad blocks (and hence,
NAND-incompatible) and (b) it does read-erase-write when you modify a
block, so it is extremely slow and does not handle power cuts at all.

Therefore,

Nack for this patch so far.

Artem.
Felix Fietkau Nov. 7, 2014, 3:20 p.m. UTC | #5
On 2014-11-07 15:59, Artem Bityutskiy wrote:
> On Wed, 2014-11-05 at 12:23 -0800, Brian Norris wrote:
>> On Wed, Sep 17, 2014 at 04:28:03PM -0400, Rodrigo Freire wrote:
>> > From: Felix Fietkau <nbd@openwrt.org>
>> > 
>> > mtd: block2mtd: Ensure that block2mtd is presented in a timely fashion
>> > 
>> > Currently, a block MTD device is not presented to the system on time, in
>> > order to start mounting the filesystems. This patch ensures that block2mtd
>> > is presented at the right time, so filesystems can be mounted on boot time.
>> > This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
>> > block2mtd filesystems.
>> 
>> This still seems like a bad idea (using a block device + block2mtd +
>> JFFS2). Why are you doing this? See comments here:
>> 
>>   http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2
> 
> At old days it was impossible to mount character devices, only block
> devices. So people ware using mtdblock to mount jffs2. This is just a
> work-around.
> 
> I did not look at this code for long time, and may have forgotten
> something, but I believe you do not need mtdblock anymore for this. You
> just mount the mtd device.
> 
> This driver should not be used for this.
> 
> The only usage of this driver is emulating a block device on top of NOR
> flash, and in most cases, only for debugging / research purposes. This
> is because (a) this driver does not handle bad blocks (and hence,
> NAND-incompatible) and (b) it does read-erase-write when you modify a
> block, so it is extremely slow and does not handle power cuts at all.
I think you got things mixed up a bit. This is not about emulating a
block device on top of NOR flash. This is about emulating NOR flash on
top of a regular block device - e.g. USB sticks, SD cards.

Now why would we want to use jffs2 on something that has enough storage
for running a regular file system?
It's simple: JFFS2 is very robust when dealing with sudden loss of
power. Block device based filesystems such as ext4, or even f2fs are not
nearly as good at dealing with that.

I realize that the emulation is a bit wasteful, and a lot of the things
that jffs2 was designed for are useless here - but the end result is
still much more reliable this way than with the alternative solutions.

In OpenWrt we've been using this approach for some x86 routers and for
the Raspberry Pi for a long time now, and it has served us quite well.

- Felix
Artem Bityutskiy Nov. 7, 2014, 3:30 p.m. UTC | #6
On Fri, 2014-11-07 at 16:20 +0100, Felix Fietkau wrote:
> > The only usage of this driver is emulating a block device on top of NOR
> > flash, and in most cases, only for debugging / research purposes. This
> > is because (a) this driver does not handle bad blocks (and hence,
> > NAND-incompatible) and (b) it does read-erase-write when you modify a
> > block, so it is extremely slow and does not handle power cuts at all.
> I think you got things mixed up a bit. This is not about emulating a
> block device on top of NOR flash. This is about emulating NOR flash on
> top of a regular block device - e.g. USB sticks, SD cards.

Gosh, I'm sorry, you are right. I did mix things up, please, ignore my
e-mail.
Rodrigo Freire Nov. 9, 2014, 12:18 p.m. UTC | #7
Hi Brian,

> From: "Brian Norris" <computersforpeace@gmail.com>
> Sent: Wednesday, November 5, 2014 6:23:03 PM

> On Wed, Sep 17, 2014 at 04:28:03PM -0400, Rodrigo Freire wrote:
> > Currently, a block MTD device is not presented to the system on time, in
> > order to start mounting the filesystems. This patch ensures that block2mtd
> > is presented at the right time, so filesystems can be mounted on boot time.
> > This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
> > block2mtd filesystems.

> This still seems like a bad idea (using a block device + block2mtd +
> JFFS2). Why are you doing this? See comments here:
> http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2

As Felix stated on a previous message to the thread, I am using JFFS2 over
block2mtd where regular filesystems failed to do so well. There are several
[1] threads pointing this issue, and JFFS2 over block2mtd works like a charm
on more harsh scenarios. The block2mtd behavior was not working correctly on
BCM2835 architecture (the kernel waited for the block device prior to its
actual enumeration by the kernel) and this patch ensures that block2mtd
kicks in _after_ the block devices was enumerated or after a user-defined
timeout. 
The patchset also enables block2mtd to define a MTD name (a MTD supports it
natively, the block2mtd had its name hard-coded to its block device name).

> You're stating right up front that this patch is doing several different
> things. Please split these up into separate commits which get their own
> description.

Done. I'll send a new split V3 patchset.

> You have several checkpatch warnings. Please fix them.

Thanks for pointing. Done.

> The addition of this name parameter should definitely be its own patch.

I agree. Done.

> This variable produces a warning when built as a module:

> drivers/mtd/devices/block2mtd.c: In function ‘add_device’:
> drivers/mtd/devices/block2mtd.c:228:6: warning: unused variable ‘i’
> [-Wunused-variable]
> int i;
> ^

Oooops. Fixed.

> > +#ifndef MODULE
> > +/*
> > +* We might not have the root device mounted at this point.
> > +* Try to resolve the device name by other means.
> > +*/

> These lines should start with a space, so the asterisks line up.

Fixed, and also fixed other non-aligned asterisks as well.

> > - dev->mtd.size = dev->blkdev->bd_inode->i_size & PAGE_MASK;
> > + dev->mtd.size = dev->blkdev->bd_inode->i_size & ~(erase_size - 1);

> This deserves its own patch, or at least some explanation of why you're
> doing this. I guess you're seeing cases where the provided erasesize is
> not aligned with the size of the block device?

Jörg pointed it on https://lkml.org/lkml/2014/9/9/680. Now, we just keep
it aligned with erase_size. Separated on a 3rd patch.

> > dev->mtd.erasesize = erase_size;
> > dev->mtd.writesize = 1;
> > dev->mtd.writebufsize = PAGE_SIZE;
> > @@ -276,15 +301,19 @@ static struct block2mtd_dev *add_device(
> > dev->mtd.priv = dev;
> > dev->mtd.owner = THIS_MODULE;
> >
> > - if (mtd_device_register(&dev->mtd, NULL, 0)) {
> > + part = kzalloc(sizeof(struct mtd_partition), GFP_KERNEL);
> > + part->name = name;
> > + part->offset = 0;
> > + part->size = dev->mtd.size;

> Why are you doing this? This also does not fit the description of this
> patch. And what's wrong with using the default partitioning options?
> Won't we (if not specified in some other way) default to an
> unpartitioned MTD, which covers the entire device?

This code was changed in order to support a MTD name.

Thanks for your dilligent review.

Best regards,

- RF.

[1] - http://bit.ly/1smGvwa
Rodrigo Freire Nov. 9, 2014, 12:18 p.m. UTC | #8
From: Felix Fietkau <nbd@openwrt.org>

mtd: block2mtd: wait for device enumeration, add name support

Currently, a block MTD device is not presented timely on boot time, in
order to start mounting the filesystems, causing the system to not boot
or panic because of lack of rootfs. This patch ensures that block2mtd
is presented at the right time, so the filesystems can be mounted on boot
time.
This issue was seen on BCM2835 (Raspberry Pi) systems when mounting JFFS2
block2mtd filesystems.
This patchset also adds a MTD device name and a timeout option to the driver
and deprecates PAGE_MASK when calculating the device size.
Original patchset:
https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/440-block2mtd_init.patch?rev=40444
https://dev.openwrt.org/browser/trunk/target/linux/generic/patches-3.12/441-block2mtd_probe.patch?rev=40444

V3: Split the changes on 3 different patches, fixes a compile warning
V2: Uses kstrdup, removed PAGE_MASK.

 drivers/mtd/devices/block2mtd.c |   57 ++++++++++++++++++++++++++++----------
 1 files changed, 42 insertions(+), 15 deletions(-)
 drivers/mtd/devices/block2mtd.c |   32 +++++++++++++++++++++++---------
 1 files changed, 23 insertions(+), 9 deletions(-)
 drivers/mtd/devices/block2mtd.c |    3 +--
 1 files changed, 1 insertions(+), 2 deletions(-)
---
1.7.1
Brian Norris Nov. 26, 2014, 3:33 a.m. UTC | #9
On Sun, Nov 09, 2014 at 07:18:05AM -0500, Rodrigo Freire wrote:
> > From: "Brian Norris" <computersforpeace@gmail.com>
> > Sent: Wednesday, November 5, 2014 6:23:03 PM
> 
> > This still seems like a bad idea (using a block device + block2mtd +
> > JFFS2). Why are you doing this? See comments here:
> > http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2
> 
> As Felix stated on a previous message to the thread, I am using JFFS2 over
> block2mtd where regular filesystems failed to do so well. There are several
> [1] threads pointing this issue, and JFFS2 over block2mtd works like a charm
> on more harsh scenarios.
[...]
> [1] - http://bit.ly/1smGvwa

OK, so there are definitely problems with cheap SD card power cut
tolerance. That's not news. But that doesn't mean block2mtd + JFFS2 is a
good solution. In fact, when I add 'jffs2' to your Google search query
of 'raspberry pi corrupt sd card', the only mentions I see are those who
agree that this is not the right choice.

But anyway, we can look at supporting block2mtd (since you provided the
patches), even if we don't agree how it should be used. And in fact, I
might argue there are no good (production) uses for block2mtd, so I
suppose I don't have much stake in it :)

Brian
Rodrigo Freire Nov. 26, 2014, 1:32 p.m. UTC | #10
From: "Brian Norris" <computersforpeace@gmail.com>
Sent: Wednesday, November 26, 2014 1:33:04 AM
Subject: Re: [PATCH V2] mtd: block2mtd: Present block2mtd timely on boot time

> On Sun, Nov 09, 2014 at 07:18:05AM -0500, Rodrigo Freire wrote:
> > > From: "Brian Norris" <computersforpeace@gmail.com>
> > > Sent: Wednesday, November 5, 2014 6:23:03 PM
> >
> > > This still seems like a bad idea (using a block device + block2mtd +
> > > JFFS2). Why are you doing this? See comments here:
> > > http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2
> >
> > As Felix stated on a previous message to the thread, I am using JFFS2 over
> > block2mtd where regular filesystems failed to do so well. There are several
> > [1] threads pointing this issue, and JFFS2 over block2mtd works like a
> > charm
> > on more harsh scenarios.
> [...]
> > [1] - http://bit.ly/1smGvwa

> OK, so there are definitely problems with cheap SD card power cut
> tolerance. That's not news. But that doesn't mean block2mtd + JFFS2 is a
> good solution. In fact, when I add 'jffs2' to your Google search query
> of 'raspberry pi corrupt sd card', the only mentions I see are those who
> agree that this is not the right choice.

> But anyway, we can look at supporting block2mtd (since you provided the
> patches), even if we don't agree how it should be used. And in fact, I
> might argue there are no good (production) uses for block2mtd, so I
> suppose I don't have much stake in it :)

Hi there Brian,

This patchset primarily aims to fix a block2mtd behavior, and not introduce new features (well, the device name and a timeout option are indeed new options, but they're actually enhancements). block2mtd already exists, works nicely as boot root device on several architectures, but fails on BCM2835 arch. Our patchset only aims to get it fixed. We just want to block2mtd work on BCM2835 the way it works on different architectures. So, this is a fix.

As a side note, WRT the SD card corruption; it also happens on good quality SD cards too. The main culprit for the corruption is bad mains / power supply issues / abrupt poweroff.

Thanks for the thorough review.

Looking forward for the ACK ;-)

My best regards,

- RF.
Rodrigo Freire Feb. 11, 2015, 3:09 p.m. UTC | #11
From: "Brian Norris" <computersforpeace@gmail.com> 
Sent: Wednesday, November 26, 2014 1:33:04 AM 
Subject: Re: [PATCH V2] mtd: block2mtd: Present block2mtd timely on boot time 

> On Sun, Nov 09, 2014 at 07:18:05AM -0500, Rodrigo Freire wrote: 
> > > From: "Brian Norris" <computersforpeace@gmail.com> 
> > > Sent: Wednesday, November 5, 2014 6:23:03 PM 
> > 
> > > This still seems like a bad idea (using a block device + block2mtd + 
> > > JFFS2). Why are you doing this? See comments here: 
> > > http://www.linux-mtd.infradead.org/faq/jffs2.html#L_hdd_jffs2 
> > 
> > As Felix stated on a previous message to the thread, I am using JFFS2 over 
> > block2mtd where regular filesystems failed to do so well. There are several
> > [1] threads pointing this issue, and JFFS2 over block2mtd works like a 
> > charm 
> > on more harsh scenarios. 
> [...] 
> > [1] - http://bit.ly/1smGvwa 

> OK, so there are definitely problems with cheap SD card power cut 
> tolerance. That's not news. But that doesn't mean block2mtd + JFFS2 is a 
> good solution. In fact, when I add 'jffs2' to your Google search query 
> of 'raspberry pi corrupt sd card', the only mentions I see are those who 
> agree that this is not the right choice. 

> But anyway, we can look at supporting block2mtd (since you provided the 
> patches), even if we don't agree how it should be used. And in fact, I 
> might argue there are no good (production) uses for block2mtd, so I 
> suppose I don't have much stake in it :) 

Hi there Brian, 

This patchset primarily aims to fix a block2mtd behavior, and not introduce
new features (well, the device name and a timeout option are indeed new
options, but they're actually enhancements). block2mtd already exists, works
 nicely as boot root device on several architectures, but fails on BCM2835
arch. Our patchset only aims to get it fixed. We just want to block2mtd work
on BCM2835 the way it works on different architectures. So, this is a fix. 

As a side note, WRT the SD card corruption; it also happens on good quality
SD cards too. The main culprit for the corruption is bad mains / power supply
issues / abrupt poweroff. And there's also the wear leveling...

Thanks for the thorough review. 

Looking forward for the ACK ;-) 

My best regards, 

- RF.
diff mbox

Patch

--- a/drivers/mtd/devices/block2mtd.c	2014-09-16 21:38:12.543952627 -0300
+++ b/drivers/mtd/devices/block2mtd.c	2014-09-17 17:43:21.424944394 -0300
@@ -9,7 +9,15 @@ 
 
 #define pr_fmt(fmt) KBUILD_MODNAME ": " fmt
 
+/*
+* When the first attempt at device initialization fails, we may need to
+* wait a little bit and retry. This timeout, by default 3 seconds, gives
+* device time to start up. Required on BCM2708 and a few other chipsets.
+*/
+#define MTD_DEFAULT_TIMEOUT	3
+
 #include <linux/module.h>
+#include <linux/delay.h>
 #include <linux/fs.h>
 #include <linux/blkdev.h>
 #include <linux/bio.h>
@@ -17,6 +25,7 @@ 
 #include <linux/list.h>
 #include <linux/init.h>
 #include <linux/mtd/mtd.h>
+#include <linux/mtd/partitions.h>
 #include <linux/mutex.h>
 #include <linux/mount.h>
 #include <linux/slab.h>
@@ -209,12 +218,14 @@  static void block2mtd_free_device(struct
 }
 
 
-static struct block2mtd_dev *add_device(char *devname, int erase_size)
+static struct block2mtd_dev *add_device(char *devname, int erase_size, const char *mtdname, int timeout)
 {
 	const fmode_t mode = FMODE_READ | FMODE_WRITE | FMODE_EXCL;
-	struct block_device *bdev;
+	struct block_device *bdev = ERR_PTR(-ENODEV);
 	struct block2mtd_dev *dev;
+	struct mtd_partition *part;
 	char *name;
+	int i;
 
 	if (!devname)
 		return NULL;
@@ -225,15 +236,28 @@  static struct block2mtd_dev *add_device(
 
 	/* Get a handle on the device */
 	bdev = blkdev_get_by_path(devname, mode, dev);
-#ifndef MODULE
-	if (IS_ERR(bdev)) {
 
-		/* We might not have rootfs mounted at this point. Try
-		   to resolve the device name by other means. */
-
-		dev_t devt = name_to_dev_t(devname);
-		if (devt)
-			bdev = blkdev_get_by_dev(devt, mode, dev);
+#ifndef MODULE
+/*
+* We might not have the root device mounted at this point.
+* Try to resolve the device name by other means.
+*/
+	for (i = 0; IS_ERR(bdev) && i <= timeout; i++) {
+		dev_t devt;
+
+		if (i)
+			/*
+			 * Calling wait_for_device_probe in the first loop 
+			 * was not enough, sleep for a bit in subsequent
+			 * go-arounds.
+			*/
+			msleep(1000);
+		wait_for_device_probe();
+
+		devt = name_to_dev_t(devname);
+		if (!devt)
+			continue;
+		bdev = blkdev_get_by_dev(devt, mode, dev);
 	}
 #endif
 
@@ -257,13 +281,14 @@  static struct block2mtd_dev *add_device(
 
 	/* Setup the MTD structure */
 	/* make the name contain the block device in */
-	name = kasprintf(GFP_KERNEL, "block2mtd: %s", devname);
+	if (!mtdname)
+		mtdname = devname;
+	name = kstrdup (mtdname, GFP_KERNEL);
 	if (!name)
 		goto err_destroy_mutex;
 
 	dev->mtd.name = name;
-
-	dev->mtd.size = dev->blkdev->bd_inode->i_size & PAGE_MASK;
+	dev->mtd.size = dev->blkdev->bd_inode->i_size & ~(erase_size - 1);
 	dev->mtd.erasesize = erase_size;
 	dev->mtd.writesize = 1;
 	dev->mtd.writebufsize = PAGE_SIZE;
@@ -276,15 +301,19 @@  static struct block2mtd_dev *add_device(
 	dev->mtd.priv = dev;
 	dev->mtd.owner = THIS_MODULE;
 
-	if (mtd_device_register(&dev->mtd, NULL, 0)) {
+	part = kzalloc(sizeof(struct mtd_partition), GFP_KERNEL);
+	part->name = name;
+	part->offset = 0;
+	part->size = dev->mtd.size;
+	if (mtd_device_register(&dev->mtd, part, 1)) {
 		/* Device didn't get added, so free the entry */
 		goto err_destroy_mutex;
 	}
+
 	list_add(&dev->list, &blkmtd_device_list);
 	pr_info("mtd%d: [%s] erase_size = %dKiB [%d]\n",
 		dev->mtd.index,
-		dev->mtd.name + strlen("block2mtd: "),
-		dev->mtd.erasesize >> 10, dev->mtd.erasesize);
+		mtdname, dev->mtd.erasesize >> 10, dev->mtd.erasesize);
 	return dev;
 
 err_destroy_mutex:
@@ -353,11 +382,12 @@  static char block2mtd_paramline[80 + 12]
 
 static int block2mtd_setup2(const char *val)
 {
-	char buf[80 + 12]; /* 80 for device, 12 for erase size */
+	char buf[80 + 12 + 80 + 8]; /* 80 for device, 12 for erase size, 80 for name, 8 for timeout */
 	char *str = buf;
-	char *token[2];
+	char *token[4];
 	char *name;
 	size_t erase_size = PAGE_SIZE;
+	unsigned long timeout = MTD_DEFAULT_TIMEOUT;
 	int i, ret;
 
 	if (strnlen(val, sizeof(buf)) >= sizeof(buf)) {
@@ -368,7 +398,7 @@  static int block2mtd_setup2(const char *
 	strcpy(str, val);
 	kill_final_newline(str);
 
-	for (i = 0; i < 2; i++)
+	for (i = 0; i < 4; i++)
 		token[i] = strsep(&str, ",");
 
 	if (str) {
@@ -395,7 +425,13 @@  static int block2mtd_setup2(const char *
 		}
 	}
 
-	add_device(name, erase_size);
+	if (token[2] && (strlen(token[2]) + 1 > 80))
+		pr_err("mtd device name too long");
+
+
+	if (token[3] && kstrtoul(token[3], 0, &timeout))
+		pr_err("invalid timeout");
+	add_device(name, erase_size, token[2], timeout);
 
 	return 0;
 }
@@ -429,7 +465,7 @@  static int block2mtd_setup(const char *v
 
 
 module_param_call(block2mtd, block2mtd_setup, NULL, NULL, 0200);
-MODULE_PARM_DESC(block2mtd, "Device to use. \"block2mtd=<dev>[,<erasesize>]\"");
+MODULE_PARM_DESC(block2mtd, "Device to use. \"block2mtd=<dev>[,<erasesize>[,<name>[,<timeout>]]]\"");
 
 static int __init block2mtd_init(void)
 {
@@ -463,8 +499,7 @@  static void block2mtd_exit(void)
 	}
 }
 
-
-module_init(block2mtd_init);
+late_initcall(block2mtd_init);
 module_exit(block2mtd_exit);
 
 MODULE_LICENSE("GPL");